CON
  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000

  'RS232
  RS232_Tx        = 30   
  RS232_Rx        = 31

  'I2C
  I2C_SDA         = 25
  I2C_SCL         = 26
  
  'CAN_INT                               
  CAN_INT_MOSI        = 17 
  CAN_INT_MISO        = 18 
  CAN_INT_CLK         = 20 
  CAN_INT_CS          = 19 
  CAN_INT_RX_INT_0    = 22 
  CAN_INT_RX_INT_1    = 23 
  CAN_INT_TX_INT_A    = 21 
  CAN_INT_Speed       = 250

  'CAN_EXT1                               
  CAN_EXT1_MOSI        = 2 
  CAN_EXT1_MISO        = 1 
  CAN_EXT1_CLK         = 3 
  CAN_EXT1_CS          = 0
  CAN_EXT1_RX_INT_0    = 5 
  CAN_EXT1_RX_INT_1    = 6 
  CAN_EXT1_TX_INT_A    = 4 
  CAN_EXT1_Speed       = 250

  'CAN_EXT2                               
  CAN_EXT2_MOSI        = 10 
  CAN_EXT2_MISO        = 9
  CAN_EXT2_CLK         = 12 
  CAN_EXT2_CS          = 8 
  CAN_EXT2_RX_INT_0    = 13 
  CAN_EXT2_RX_INT_1    = 14 
  CAN_EXT2_TX_INT_A    = 11 
  CAN_EXT2_Speed       = 250
                                            
  'EEPROM BaseAdress = $8000
  Adress_NodeID   = $8000                                                                     
  Adress_Serial   = $8002
  Adress_Interval = $8004
  Adress_Frequency= $8006
  
  'MISC 
  LED        = 15
  ADC_Addr   = %000
                                                                  
VAR

  'SYSTEM
  long Type 
  long Baudrate
  long Counter    'Trigger for Heartbeat
  long Accepted
  long Version  
  long Status
  long Pause
  long ValueChangeRequested '-1 if no request pending else ID of value
  long LastCnt
  long CurrentCnt
  byte RomReadBuffer[32]
  long CommunicationHalt

  'EEPROM
  long Node_ID  
  long Serial
  long Interval
  
  'INPUTS
  long ADC_Value[4] 

  'CAN_INT
  long ID
  long ByteCnt
  byte DataBuff[8]
  byte RxMsgBuffer[16]
  byte TxMsgBuffer[16]
  long TxMsgCounter 'counts which type of message is to be sent at next interval
  long Int_MsgToSend_New
  long Int_MsgToSend_ID
  long Int_MsgToSend_ByteCnt 
  long Int_MsgToSend_Databuffer[8]
  
  'CAN_EXT1
  long Ext1_ID
  long Ext1_ByteCnt
  byte Ext1_DataBuff[8]
  byte Ext1_RxMsgBuffer[16]
  byte Ext1_TxMsgBuffer[16]
  long Ext1_MsgToSend_New 
  long Ext1_MsgToSend_ID
  long Ext1_MsgToSend_ByteCnt 
  long Ext1_MsgToSend_Databuffer[8]
   
  'CAN_EXT2
  long Ext2_ID
  long Ext2_ByteCnt
  byte Ext2_DataBuff[8]
  byte Ext2_RxMsgBuffer[16]
  byte Ext2_TxMsgBuffer[16]
  long Ext2_MsgToSend_New 
  long Ext2_MsgToSend_ID
  long Ext2_MsgToSend_ByteCnt 
  long Ext2_MsgToSend_Databuffer[8]
  
  'COG_RAM
  long Stack1[500]
  long Stack2[500]
  long Stack3[500]
  long Stack4[500]
  
  'MISC
  long LoopIntact '1 if heartbeat from primary is received by secondary, 0 if not. if loop is intact, messages will only be forwarded to primary, else they will be sent on both
  long LoopLastCnt
  long LoopCurrentCnt     
  long LoopCurrentCntTick 'Ticks one up each 80MCyc
  
OBJ

  Functions     :    "100001.Functions"
  Can1          :    "100001.Driver_MCP2515_PacketHandler"
  Can2          :    "100001.Driver_MCP2515_PacketHandler"
  Can3          :    "100001.Driver_MCP2515_PacketHandler"
  EEPROM        :    "100001.Driver_EEPROM_512K"
  ADC           :    "100001.Driver_MCP3428"
  
PUB main | i 

  'SYSTEM
  Type := 8                                          'Type of product, ref. CAN-Bus Communication protocol
  Counter := 0                                       'Counter for heartbeat
  Accepted := 0                                      '0=not accepted, 1=accepted by host
  Version := 1                                       'Software version number
  ValueChangeRequested := -1                         'Set pending change request to false
  dira[LED]~~

  'Initiate Loop timeout detect on CAN-Bus
  LoopIntact := 1 'Assume loop is OK on startup 
  LoopLastCnt := cnt
  
  'CAN_START
  Can1.Start(CAN_INT_CS, CAN_INT_MISO, CAN_INT_MOSI, CAN_INT_CLK, CAN_INT_TX_INT_A, CAN_INT_RX_INT_0, CAN_INT_RX_INT_1, CAN_INT_Speed)
  Can2.Start(CAN_EXT1_CS, CAN_EXT1_MISO, CAN_EXT1_MOSI, CAN_EXT1_CLK, CAN_EXT1_TX_INT_A, CAN_EXT1_RX_INT_0, CAN_EXT1_RX_INT_1, CAN_EXT1_Speed)
  Can3.Start(CAN_EXT2_CS, CAN_EXT2_MISO, CAN_EXT2_MOSI, CAN_EXT2_CLK, CAN_EXT2_TX_INT_A, CAN_EXT2_RX_INT_0, CAN_EXT2_RX_INT_1, CAN_EXT2_Speed)
    
  'EEPROM
  EEPROM.Start(%000)
  Node_ID := EEPROM.rd_byte(Adress_NodeID)                     'CAN_ID 
  Serial := EEPROM.rd_word(Adress_Serial)                      'Serial number
  Interval := ((EEPROM.rd_word(Adress_Interval) #> 1) <# 3000) 'Sleep time in ms

  'COG_START
  cognew(Receive_CAN, @Stack1)
  cognew(Receive_CAN_EXT1, @Stack2)
  cognew(Receive_CAN_EXT2, @Stack3)
  cognew(TransmitCog, @Stack4)
  
  Adc.Start(I2C_SDA,I2C_SCL)
 
  repeat
    repeat while CommunicationHalt == 1
  
    if Accepted == 1
      ADC_Read   
      Transmit_CAN
    !outa[LED]   

    Pause := 80_000 * Interval        
    Counter++
    if Counter => ((80_000_000 / Pause) * 3) 
      Transmit_Heartbeat
      Counter := 0
    waitcnt(Pause + cnt)
     
PUB ADC_Read | i

  repeat i from 0 to 3                     'Loop through channels one to four on ADC
    ADC.Request_Read(ADC_Addr, i)          'Prepare measurement in ADC for current channel
    waitcnt(500_000 + cnt)                 'Wait for measurements to be ready
    ADC_Value[i] := ADC.Read(ADC_Addr)     'Read requested channel from ADC

PUB TransmitCog

  repeat 

    if Int_MsgToSend_New == 2
      Can1.EnquePacket(Int_MsgToSend_ID, Int_MsgToSend_ByteCnt, @Int_MsgToSend_Databuffer)
      Int_MsgToSend_New := 0
      
    if Ext1_MsgToSend_New == 2
      Can2.EnquePacket(Ext1_MsgToSend_ID, Ext1_MsgToSend_ByteCnt, @Ext1_MsgToSend_Databuffer)
      Ext1_MsgToSend_New := 0

    if Ext2_MsgToSend_New == 2
      Can3.EnquePacket(Ext2_MsgToSend_ID, Ext2_MsgToSend_ByteCnt, @Ext2_MsgToSend_Databuffer)
      Ext2_MsgToSend_New := 0

PUB Transmit_CAN | Can_Tx 

  TxMsgCounter++
  if TxMsgCounter == 2
    TxMsgCounter := 0
  
  case TxMsgCounter

    0:  
      Can_Tx := Type                       'Type           
      DataBuff[0] := Can_Tx.byte[0]
      Can_Tx := Node_ID                    'Node_ID
      DataBuff[1] := Can_Tx.byte[0]
      Can_Tx := 0                          'StartVar       
      DataBuff[2] := Can_Tx.byte[0]
      Can_Tx := 1                          'OutType
      DataBuff[3] := Can_Tx.byte[0]
      Can_Tx := ADC_Value[0]               'Current 5v  
      DataBuff[4] := Can_Tx.byte[0]
      DataBuff[5] := Can_Tx.byte[1] 
      Can_Tx := (ADC_Value[1] * 110) / 10 'Voltage 5v   
      DataBuff[6] := Can_Tx.byte[0]                             
      DataBuff[7] := Can_Tx.byte[1]
      PrepareCanMsg(0, 20, 8, @DataBuff)
      
    1:
      Can_Tx := Type                       'Type           
      DataBuff[0] := Can_Tx.byte[0]
      Can_Tx := Node_ID                    'Node_ID
      DataBuff[1] := Can_Tx.byte[0]
      Can_Tx := 2                          'StartVar       
      DataBuff[2] := Can_Tx.byte[0]
      Can_Tx := 1                          'OutType
      DataBuff[3] := Can_Tx.byte[0]
      Can_Tx := ADC_Value[2]               'Current 24v  
      DataBuff[4] := Can_Tx.byte[0]
      DataBuff[5] := Can_Tx.byte[1] 
      Can_Tx := (ADC_Value[3] * 1614) / 100 'Voltage 24v   
      DataBuff[6] := Can_Tx.byte[0] 
      DataBuff[7] := Can_Tx.byte[1]
      PrepareCanMsg(0, 20, 8, @DataBuff)
    
PUB Transmit_Heartbeat 

  Status := 0 
  if ValueChangeRequested > -1
    Status |= %0000_0100 
  if Accepted == 0
    Status |= %0000_1000 

  DataBuff[0] := Type
  DataBuff[1] := Node_ID
  DataBuff[2] := Serial.byte[0]  
  DataBuff[3] := Serial.byte[1]   
  DataBuff[4] := Status
  DataBuff[5] := 0
  DataBuff[6] := 0
  DataBuff[7] := Version
  PrepareCanMsg(0, 1, 8, @DataBuff)    'Heartbeat on Internal  
  PrepareCanMsg(1, 1, 8, @DataBuff)    'Heartbeat sent on primary to detect loop to secondary 

PUB PrepareCanMsg(Can_No, Can_ID, Can_ByteCnt, DatabufferPtr)

  case Can_No
    0:
      repeat while Int_MsgToSend_New > 0
      Int_MsgToSend_New     := 1
      bytemove(@Int_MsgToSend_Databuffer, DatabufferPtr, 8) 'Copy buffer to prepare response     
      Int_MsgToSend_ID      := Can_ID
      Int_MsgToSend_ByteCnt := Can_ByteCnt
      Int_MsgToSend_New     := 2
    1:
      repeat while Ext1_MsgToSend_New > 0
      Ext1_MsgToSend_New     := 1
      bytemove(@Ext1_MsgToSend_Databuffer, DatabufferPtr, 8) 'Copy buffer to prepare response     
      Ext1_MsgToSend_ID      := Can_ID
      Ext1_MsgToSend_ByteCnt := Can_ByteCnt
      Ext1_MsgToSend_New     := 2       
    2:
      repeat while Ext2_MsgToSend_New > 0
      Ext2_MsgToSend_New     := 1
      bytemove(@Ext2_MsgToSend_Databuffer, DatabufferPtr, 8) 'Copy buffer to prepare response     
      Ext2_MsgToSend_ID      := Can_ID
      Ext2_MsgToSend_ByteCnt := Can_ByteCnt
      Ext2_MsgToSend_New     := 2
 
PUB Receive_CAN | i, a, nMsgId, Checksum, TempLong

  'Set interupt pins to Inputs
  dira[CAN_INT_RX_INT_0]~
  dira[CAN_INT_RX_INT_1]~
  
  repeat

    'Check interrupts for Rx0Int or Rx1Int and Read Receive Buffer in MCP2515  
    if ina[CAN_INT_RX_INT_0] == 0
      Can1.ReadReceiveBuffer(0, @RxMsgBuffer)
    elseif ina[CAN_INT_RX_INT_1] == 0
      Can1.ReadReceiveBuffer(1, @RxMsgBuffer)
    else
      bytefill(@RxMsgBuffer, 0, 16)
      
    'Get ID
    nMsgId := RxMsgBuffer[0] << 3
    nMsgId |= (RxMsgBuffer[1] >> 5) & $07

    'Forward message
    if ((nMsgId > 0) or (RxMsgBuffer[5] == 255))     'Verify there is a message to send
      if !((RxMsgBuffer[5] == Type) and (RxMsgBuffer[6] == Node_ID))
        PrepareCanMsg(1, nMsgId, 8, @RxMsgBuffer[5])   'Send on Primary
        if LoopIntact == 0                             'If loop broken
          PrepareCanMsg(2, nMsgId, 8, @RxMsgBuffer[5]) 'Send on Secondary also 
        
    'Check ID  of incoming message
    Case nMsgId
      
      2: 'Node accepted or rejected
        if RxMsgBuffer[5] == Type
          if RxMsgBuffer[6] == Node_ID
            if RxMsgBuffer[7] == 1
              Accepted := 1
            else
              Accepted := 0
       
      3: 'Request Read or Set Value      
        if RxMsgBuffer[5] == Type
          if RxMsgBuffer[6] == Node_ID
            if ((RxMsgBuffer[8] * 256) + RxMsgBuffer[7]) == Serial
              if RxMsgBuffer[9] == 0      'Check OP-Code for settings read request  
                bytemove(@TxMsgBuffer, @RxMsgBuffer, 16) 'Copy buffer to prepare response
                TxMsgBuffer[9] := 1       'OP-Code Read Response
                TxMsgBuffer[10] := EEPROM.rd_byte($8000 + (((RxMsgBuffer[11] * 256) + RxMsgBuffer[10]) * 2)) 
                TxMsgBuffer[11] := EEPROM.rd_byte($8001 + (((RxMsgBuffer[11] * 256) + RxMsgBuffer[10]) * 2))
                PrepareCanMsg(0, nMsgId, 8, @TxMsgBuffer[5])
              elseif RxMsgBuffer[9] == 2  'Check OP-Code for settings read request   
                ValueChangeRequested := ((RxMsgBuffer[11] * 256) + RxMsgBuffer[10]) 'Store variable to be changed
                TxMsgBuffer[9] := 3       'OP-Code Set Value Response
                PrepareCanMsg(0, nMsgId, 8, @TxMsgBuffer[5])   
                
      4: 'Change in Setting
        if ValueChangeRequested > -1
          if RxMsgBuffer[5] == Type
            if RxMsgBuffer[6] == Node_ID
              if ((RxMsgBuffer[8] * 256) + RxMsgBuffer[7]) == Serial           
                if RxMsgBuffer[9] == 2  'Check OP-Code for settings change request
                  bytemove(@TxMsgBuffer, @RxMsgBuffer,16) 'Copy buffer to prepare response 
                  SaveSetting(ValueChangeRequested, (RxMsgBuffer[11] * 256) + RxMsgBuffer[10])
                  TxMsgBuffer[9] := 3   'Set OP-Code to Answer
                  PrepareCanMsg(0, nMsgId, 8, @TxMsgBuffer[5])
                  ValueChangeRequested := -1

      6: 'Communication Halt
        CommunicationHalt := (RxMsgBuffer[7] & $01)
    
      7: 'Reboot
        if RxMsgBuffer[5] == Type
          if RxMsgBuffer[6] == Node_ID
            Reboot
        if RxMsgBuffer[7] == 1    'Reboot All
          Reboot

      8: 'Get Long from EEPROM, used for downloading software from propeller
        if RxMsgBuffer[5] == Type
          if RxMsgBuffer[6] == Node_ID
            bytemove(@TxMsgBuffer, @RxMsgBuffer, 16) 'Copy buffer to prepare response    
            TempLong := EEPROM.rd_long(((RxMsgBuffer[8] * 256) + RxMsgBuffer[7]))
            repeat i from 0 to 3
              TxMsgBuffer[9 + i] := TempLong.byte[0 + i]
            PrepareCanMsg(0, nMsgId, 8, @TxMsgBuffer[5])
      
      9: 'Set Long in EEPROM, used for updating firmware to propeller
        if (RxMsgBuffer[5] & $FF) == Type
          if RxMsgBuffer[6] == Node_ID
            bytemove(@TxMsgBuffer, @RxMsgBuffer, 16) 'Copy buffer to prepare response    
            EEPROM.wr_block(((RxMsgBuffer[8] * 256) + RxMsgBuffer[7]), 4, @RxMsgBuffer[9])
            PrepareCanMsg(0, nMsgId, 8, @TxMsgBuffer[5]) 
            
      10: 'Get checksum of first 32KB in EEPROM
        if RxMsgBuffer[5] == Type
          if RxMsgBuffer[6] == Node_ID
            bytemove(@TxMsgBuffer, @RxMsgBuffer, 16) 'Copy buffer to prepare response            
            Checksum := 0
            repeat i from 0 to 1023
              EEPROM.rd_block(i * 32, 32, @RomReadBuffer[0])
              repeat a from 0 to 31
                Checksum ^= RomReadBuffer[a]
            TxMsgBuffer[7] := Checksum
            PrepareCanMsg(0, nMsgId, 8, @TxMsgBuffer[5]) 


PUB SaveSetting(Settings_ID, Settings_Value)

  'Write new setting to EEPROM
  EEPROM.wr_word($8000 + (Settings_ID * 2), Settings_Value)

  'Update variable in use
  Case Settings_ID
    0:
      Node_ID := Settings_Value
    1:
      Serial := Settings_Value
    2:
      Interval := Settings_Value
   
PUB Receive_CAN_EXT1 | i, a, nMsgId, Checksum, TempLong

  'Set interupt pins to Inputs
  dira[CAN_EXT1_RX_INT_0]~
  dira[CAN_EXT1_RX_INT_1]~
  
  repeat

    'Check interrupts for Rx0Int or Rx1Int and Read Receive Buffer in MCP2515  
    if ina[CAN_EXT1_RX_INT_0] == 0
      Can2.ReadReceiveBuffer(0, @Ext1_RxMsgBuffer)
    elseif ina[CAN_EXT1_RX_INT_1] == 0
      Can2.ReadReceiveBuffer(1, @Ext1_RxMsgBuffer)
    else
      bytefill(@Ext1_RxMsgBuffer, 0, 16)
      
    'Get ID
    nMsgId := Ext1_RxMsgBuffer[0] << 3
    nMsgId |= (Ext1_RxMsgBuffer[1] >> 5) & $07

    'Forward message
    if nMsgId > 0
      if !((RxMsgBuffer[5] == Type) and (RxMsgBuffer[6] == Node_ID))
        'Forward message on Internal CAN
        PrepareCanMsg(0, nMsgId, 8, @EXT1_RxMsgBuffer[5])       
    
    'Check ID  of incoming message
    Case nMsgId
      
      2: 'Node accepted or rejected
        if Ext1_RxMsgBuffer[5] == Type
          if Ext1_RxMsgBuffer[6] == Node_ID
            if Ext1_RxMsgBuffer[7] == 1                                  
              Accepted := 1
            else
              Accepted := 0
       
      3: 'Request Read or Set Value      
        if Ext1_RxMsgBuffer[5] == Type
          if Ext1_RxMsgBuffer[6] == Node_ID
            if ((Ext1_RxMsgBuffer[8] * 256) + Ext1_RxMsgBuffer[7]) == Serial
              if Ext1_RxMsgBuffer[9] == 0      'Check OP-Code for settings read request
                bytemove(@Ext1_TxMsgBuffer, @Ext1_RxMsgBuffer, 16) 'Copy buffer to prepare response
                Ext1_TxMsgBuffer[9] := 1       'OP-Code Read Response
                Ext1_TxMsgBuffer[10] := EEPROM.rd_byte($8000 + (((Ext1_RxMsgBuffer[11] * 256) + Ext1_RxMsgBuffer[10]) * 2)) 
                Ext1_TxMsgBuffer[11] := EEPROM.rd_byte($8001 + (((Ext1_RxMsgBuffer[11] * 256) + Ext1_RxMsgBuffer[10]) * 2))
                PrepareCanMsg(1, nMsgId, 8, @Ext1_TxMsgBuffer[5]) 
              elseif Ext1_RxMsgBuffer[9] == 2  'Check OP-Code for settings read request   
                ValueChangeRequested := ((Ext1_RxMsgBuffer[11] * 256) + Ext1_RxMsgBuffer[10]) 'Store variable to be changed
                Ext1_TxMsgBuffer[9] := 3       'OP-Code Set Value Response
                PrepareCanMsg(1, nMsgId, 8, @Ext1_TxMsgBuffer[5])  
                
      4: 'Change in Setting
        if ValueChangeRequested > -1
          if Ext1_RxMsgBuffer[5] == Type
            if Ext1_RxMsgBuffer[6] == Node_ID
              if ((Ext1_RxMsgBuffer[8] * 256) + Ext1_RxMsgBuffer[7]) == Serial           
                if Ext1_RxMsgBuffer[9] == 2  'Check OP-Code for settings change request
                  bytemove(@Ext1_TxMsgBuffer, @Ext1_RxMsgBuffer,16) 'Copy buffer to prepare response 
                  SaveSetting(ValueChangeRequested, (Ext1_RxMsgBuffer[11] * 256) + Ext1_RxMsgBuffer[10])
                  Ext1_TxMsgBuffer[9] := 3   'Set OP-Code to Answer
                  PrepareCanMsg(1, nMsgId, 8, @Ext1_TxMsgBuffer[5])  
                  ValueChangeRequested := -1
                  
      6: 'Communication Halt
        CommunicationHalt := (Ext1_RxMsgBuffer[7] & $01)
    
      7: 'Reboot
        if Ext1_RxMsgBuffer[5] == Type
          if Ext1_RxMsgBuffer[6] == Node_ID
            Reboot
        if Ext1_RxMsgBuffer[7] == 1    'Reboot All
          Reboot

      8: 'Get Long from EEPROM, used for downloading software from propeller
        if Ext1_RxMsgBuffer[5] == Type
          if Ext1_RxMsgBuffer[6] == Node_ID
            bytemove(@Ext1_TxMsgBuffer, @Ext1_RxMsgBuffer, 16) 'Copy buffer to prepare response    
            TempLong := EEPROM.rd_long(((Ext1_RxMsgBuffer[8] * 256) + Ext1_RxMsgBuffer[7]))
            repeat i from 0 to 3
              Ext1_TxMsgBuffer[9 + i] := TempLong.byte[0 + i]
            PrepareCanMsg(1, nMsgId, 8, @Ext1_TxMsgBuffer[5])  
      
      9: 'Set Long in EEPROM, used for updating firmware to propeller
        if Ext1_RxMsgBuffer[5] == Type
          if Ext1_RxMsgBuffer[6] == Node_ID
            bytemove(@Ext1_TxMsgBuffer, @Ext1_RxMsgBuffer, 16) 'Copy buffer to prepare response    
            EEPROM.wr_block(((Ext1_RxMsgBuffer[8] * 256) + Ext1_RxMsgBuffer[7]), 4, @Ext1_RxMsgBuffer[9])
            PrepareCanMsg(1, nMsgId, 8, @Ext1_TxMsgBuffer[5])  
            
      10: 'Get checksum of first 32KB in EEPROM
        if Ext1_RxMsgBuffer[5] == Type
          if Ext1_RxMsgBuffer[6] == Node_ID
            bytemove(@Ext1_TxMsgBuffer, @Ext1_RxMsgBuffer, 16) 'Copy buffer to prepare response            
            Checksum := 0
            repeat i from 0 to 1023
              EEPROM.rd_block(i * 32, 32, @RomReadBuffer[0])
              repeat a from 0 to 31
                Checksum ^= RomReadBuffer[a]
            Ext1_TxMsgBuffer[7] := Checksum
            PrepareCanMsg(1, nMsgId, 8, @Ext1_TxMsgBuffer[5])

PUB Receive_CAN_EXT2 | i, a, nMsgId, Checksum, TempLong

  'Set interupt pins to Inputs
  dira[CAN_EXT2_RX_INT_0]~
  dira[CAN_EXT2_RX_INT_1]~

  repeat

    LoopCurrentCnt := (Cnt - LoopLastCnt)  'Timeout check for loopfailure on CAN-Bus
    if LoopCurrentCnt > 80_000_000         'Check if one second has passed without heartbeat
      LoopLastCnt := cnt                   'Reset lastcnt
      LoopCurrentCntTick++                 'Tick second counter since last heartbeat
      if LoopCurrentCntTick > 10           'If ten seconds has passed
        LoopCurrentCntTick := 10           'Limit variable for running off
        LoopIntact := 0                    'Set loopintact to NO
  
    'Check interrupts for Rx0Int or Rx1Int and Read Receive Buffer in MCP2515  
    if ina[CAN_EXT2_RX_INT_0] == 0
      Can3.ReadReceiveBuffer(0, @EXT2_RxMsgBuffer)
    elseif ina[CAN_EXT2_RX_INT_1] == 0
      Can3.ReadReceiveBuffer(1, @EXT2_RxMsgBuffer)
    else
      bytefill(@EXT2_RxMsgBuffer, 0, 16)
      
    'Get ID
    nMsgId := EXT2_RxMsgBuffer[0] << 3
    nMsgId |= (EXT2_RxMsgBuffer[1] >> 5) & $07

    if ((nMsgId == 1) and (EXT2_RxMsgBuffer[5] == Type) and (EXT2_RxMsgBuffer[6] == Node_ID))
      LoopCurrentCntTick := 0
      LoopLastCnt := cnt
      LoopIntact := 1
    
    elseif (LoopIntact == 0) and (nMsgId > 0)
      if !((RxMsgBuffer[5] == Type) and (RxMsgBuffer[6] == Node_ID))
        bytemove(@EXT2_TxMsgBuffer, @EXT2_RxMsgBuffer, 16) 'Copy buffer to prepare response
        if nMsgId == 1
          EXT2_TxMsgBuffer[9] |= %0001_0000
        PrepareCanMsg(0, nMsgId, 8, @EXT2_TxMsgBuffer[5]) 'Forward message on Internal CAN
        PrepareCanMsg(1, nMsgId, 8, @EXT2_TxMsgBuffer[5]) 'Forward message on Primary CAN
      
      'Check ID  of incoming message
      Case nMsgId
        
        2: 'Node accepted or rejected
          if EXT2_RxMsgBuffer[5] == Type
            if EXT2_RxMsgBuffer[6] == Node_ID
              if EXT2_RxMsgBuffer[7] == 1
                Accepted := 1
              else
                Accepted := 0
         
        3: 'Request Read or Set Value      
          if EXT2_RxMsgBuffer[5] == Type
            if EXT2_RxMsgBuffer[6] == Node_ID
              if ((EXT2_RxMsgBuffer[8] * 256) + EXT2_RxMsgBuffer[7]) == Serial
                if EXT2_RxMsgBuffer[9] == 0      'Check OP-Code for settings read request
                  bytemove(@EXT2_TxMsgBuffer, @EXT2_RxMsgBuffer, 16) 'Copy buffer to prepare response
                  EXT2_TxMsgBuffer[9] := 1       'OP-Code Read Response
                  EXT2_TxMsgBuffer[10] := EEPROM.rd_byte($8000 + (((EXT2_RxMsgBuffer[11] * 256) + EXT2_RxMsgBuffer[10]) * 2)) 
                  EXT2_TxMsgBuffer[11] := EEPROM.rd_byte($8001 + (((EXT2_RxMsgBuffer[11] * 256) + EXT2_RxMsgBuffer[10]) * 2))
                  PrepareCanMsg(2, nMsgId, 8, @Ext2_TxMsgBuffer[5])  
                elseif EXT2_RxMsgBuffer[9] == 2  'Check OP-Code for settings read request   
                  ValueChangeRequested := ((EXT2_RxMsgBuffer[11] * 256) + EXT2_RxMsgBuffer[10]) 'Store variable to be changed
                  EXT2_TxMsgBuffer[9] := 3       'OP-Code Set Value Response
                  PrepareCanMsg(2, nMsgId, 8, @Ext2_TxMsgBuffer[5]) 
                  
        4: 'Change in Setting
          if ValueChangeRequested > -1
            if EXT2_RxMsgBuffer[5] == Type
              if EXT2_RxMsgBuffer[6] == Node_ID
                if ((EXT2_RxMsgBuffer[8] * 256) + EXT2_RxMsgBuffer[7]) == Serial           
                  if EXT2_RxMsgBuffer[9] == 2  'Check OP-Code for settings change request
                    bytemove(@EXT2_TxMsgBuffer, @EXT2_RxMsgBuffer,16) 'Copy buffer to prepare response 
                    SaveSetting(ValueChangeRequested, (EXT2_RxMsgBuffer[11] * 256) + EXT2_RxMsgBuffer[10])
                    EXT2_TxMsgBuffer[9] := 3   'Set OP-Code to Answer
                    PrepareCanMsg(2, nMsgId, 8, @Ext2_TxMsgBuffer[5]) 
                    ValueChangeRequested := -1

        6: 'Communication Halt
          CommunicationHalt := (EXT2_RxMsgBuffer[7] & $01)
        
        7: 'Reboot
          if EXT2_RxMsgBuffer[5] == Type
            if EXT2_RxMsgBuffer[6] == Node_ID
              Reboot
          if EXT2_RxMsgBuffer[7] == 1    'Reboot All
            Reboot
        
        8: 'Get Long from EEPROM, used for downloading software from propeller
          if EXT2_RxMsgBuffer[5] == Type
            if EXT2_RxMsgBuffer[6] == Node_ID
              bytemove(@EXT2_TxMsgBuffer, @EXT2_RxMsgBuffer, 16) 'Copy buffer to prepare response    
              TempLong := EEPROM.rd_long(((EXT2_RxMsgBuffer[8] * 256) + EXT2_RxMsgBuffer[7]))
              repeat i from 0 to 3
                EXT2_TxMsgBuffer[9 + i] := TempLong.byte[0 + i]
              PrepareCanMsg(2, nMsgId, 8, @Ext2_TxMsgBuffer[5])
        
        9: 'Set Long in EEPROM, used for updating firmware to propeller
          if EXT2_RxMsgBuffer[5] == Type
            if EXT2_RxMsgBuffer[6] == Node_ID
              bytemove(@EXT2_TxMsgBuffer, @EXT2_RxMsgBuffer, 16) 'Copy buffer to prepare response    
              EEPROM.wr_block(((EXT2_RxMsgBuffer[8] * 256) + EXT2_RxMsgBuffer[7]), 4, @EXT2_RxMsgBuffer[9])
              PrepareCanMsg(2, nMsgId, 8, @Ext2_TxMsgBuffer[5]) 
              
        10: 'Get checksum of first 32KB in EEPROM
          if EXT2_RxMsgBuffer[5] == Type
            if EXT2_RxMsgBuffer[6] == Node_ID
              bytemove(@EXT2_TxMsgBuffer, @EXT2_RxMsgBuffer, 16) 'Copy buffer to prepare response            
              Checksum := 0
              repeat i from 0 to 1023
                EEPROM.rd_block(i * 32, 32, @RomReadBuffer[0])
                repeat a from 0 to 31
                  Checksum ^= RomReadBuffer[a]
              EXT2_TxMsgBuffer[7] := Checksum
              PrepareCanMsg(2, nMsgId, 8, @Ext2_TxMsgBuffer[5]) 
         
            